// Copyright (c) 2014-2019 Sombusta, All Rights Reserved.
// MIT LICENSE // Epic VR Template Convert C++ Open Source Project.

#include "TP_MotionController.h"
#include "TP_InteractionInterface.h"
#include "TP_HandAnimInstance.h"
#include "UObject/ConstructorHelpers.h"
#include "Engine/World.h"
#include "Kismet/GameplayStatics.h"
#include "Kismet/KismetMathLibrary.h"
#include "Kismet/KismetSystemLibrary.h"
#include "HeadMountedDisplayFunctionLibrary.h"
#include "NavigationSystem.h"
#include "Haptics/HapticFeedbackEffect_Base.h"
#include "Components/SceneComponent.h"
#include "Components/ArrowComponent.h"
#include "Components/SphereComponent.h"
#include "Components/SplineComponent.h"
#include "Components/SplineMeshComponent.h"
#include "Components/StaticMeshComponent.h"
#include "Components/SkeletalMeshComponent.h"
#include "Components/PrimitiveComponent.h"
#include "MotionControllerComponent.h"
#include "SteamVRChaperoneComponent.h"
#include "Engine/StaticMesh.h"
#include "Engine/SkeletalMesh.h"
#include "Materials/Material.h"
#include "Macros/Macros.h"
#include "Materials/MaterialInstanceConstant.h"
//#include "Engine/Plugins/FX/Niagara/Source/NiagaraEditor/Private/ViewModels/Stack/NiagaraStackModuleItem.cpp"

class WidgetInteractionBlueprint;

// Sets default values
ATP_MotionController::ATP_MotionController()
{
	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

	// Object Initialize	
	//=============// HandMesh->SetAnimInstanceClass(AnimBP_HandAnimation.Object->GeneratedClass);

	// Animation Blueprint // cast "UClass" and Asset path is Different Why? -> Because UAnimBlueprint Class will crash a packaged game. so use postfix "_C", animation blueprint cast to UClass.
	static ConstructorHelpers::FObjectFinder<UClass> AnimBP_HandAnimation(TEXT("Class'/Game/VirtualReality/Mannequin/Animations/AnimBP_SomRightHand.AnimBP_SomRightHand_C'"));

	static ConstructorHelpers::FObjectFinder<USkeletalMesh> SK_Hand(TEXT("SkeletalMesh'/Game/VirtualReality/Mannequin/Character/Mesh/MannequinHand_Right.MannequinHand_Right'"));
	static ConstructorHelpers::FObjectFinder<UStaticMesh> SM_Sphere(TEXT("StaticMesh'/Engine/BasicShapes/Sphere.Sphere'"));
	static ConstructorHelpers::FObjectFinder<UStaticMesh> SM_Cylinder(TEXT("StaticMesh'/Engine/BasicShapes/Cylinder.Cylinder'"));
	static ConstructorHelpers::FObjectFinder<UMaterialInstanceConstant> MI_ChaperoneOutline(TEXT("MaterialInstanceConstant'/Game/VirtualReality/Materials/MI_ChaperoneOutline.MI_ChaperoneOutline'"));
	static ConstructorHelpers::FObjectFinder<UHapticFeedbackEffect_Base> HapticEffect(TEXT("HapticFeedbackEffect_Curve'/Game/VirtualRealityCPP/Blueprints/MotionControllerHaptics.MotionControllerHaptics'"));

	// Create Components Initialize
	RootScene = CreateDefaultSubobject<USceneComponent>(TEXT("RootScene"));
	MotionController = CreateDefaultSubobject<UMotionControllerComponent>(TEXT("MotionContoller"));
	HandMesh = CreateDefaultSubobject<USkeletalMeshComponent>(TEXT("HandMesh"));
	GrabShpere = CreateDefaultSubobject<USphereComponent>(TEXT("GrabShpere"));
	SteamVRChaperone = CreateDefaultSubobject<USteamVRChaperoneComponent>(TEXT("SteamVRChaperone"));

	// Setup RootComponent
	RootComponent = RootScene;

	// Setup Components
	HandMesh->SetupAttachment(MotionController);
	HandMesh->SetRelativeLocation(FVector(-12.8f, -0.03f, -1.8f));
	HandMesh->SetRelativeRotation(FRotator(0.0f, 0.0f, 90.0f));
	HandMesh->SetCollisionProfileName(TEXT("NoCollision"));
	HandMesh->CastShadow = false;
	if (SK_Hand.Succeeded())
	{
		HandMesh->SetSkeletalMesh(SK_Hand.Object);
		HandMesh->SetAnimationMode(EAnimationMode::AnimationBlueprint);
		if (AnimBP_HandAnimation.Succeeded())
		{
			HandMesh->SetAnimInstanceClass(AnimBP_HandAnimation.Object);
		}
	}

	MotionController->SetupAttachment(RootComponent);
	MotionController->SetRelativeLocation(FVector(0.0f, 0.0f, 0.0f));
	MotionController->SetCollisionProfileName(TEXT("BlockAll"));

	GrabShpere->SetupAttachment(HandMesh);
	GrabShpere->SetRelativeLocation(FVector(14.286342f, 0.224237f, 1.484932f));
	GrabShpere->SetSphereRadius(10.0f);
	GrabShpere->SetCollisionProfileName(TEXT("OverlapAllDynamic"));
	GrabShpere->SetGenerateOverlapEvents(true);
	GrabShpere->OnComponentBeginOverlap.AddDynamic(this, &ATP_MotionController::OnComponentBeginOverlap);

	if (HapticEffect.Succeeded())
	{
		VRHapticEffect = HapticEffect.Object;
	}

	Hand = EControllerHand::Left;
	AttachedActor = nullptr;
	bIsRoomScale = false;
	CurrentGripState = EGrip_Code::Open;
	bWantsToGrip = false;
	InitialControllerRotation = FRotator(0.0f, 0.0f, 0.0f);

	GetMotionControllerMeshes();
	MotionControllerMesh = CreateDefaultSubobject<UStaticMeshComponent>(TEXT("MotionControllerStaticMesh"));
	MotionControllerMesh->SetupAttachment(HandMesh);
	

}



void ATP_MotionController::OnConstruction(const FTransform & Transform)
{
	Super::OnConstruction(Transform);

	// Deprecated Hand Variable in 4.19
	// MotionController->Hand = Hand; // 4.18 Code
	MotionController->SetTrackingSource(Hand); // 4.19 Code
}

// Called when the game starts or when spawned
void ATP_MotionController::BeginPlay()
{
	Super::BeginPlay();
	SpawnWidget();


	bool IsRightHand = (Hand == EControllerHand::Right);
	if (!IsRightHand) {
		HandMesh->SetWorldScale3D(FVector(1.0f, 1.0f, -1.0f));
	}


	UWorld* MyWorld = GetWorld();
	FString CurrentMapName = MyWorld->GetName();
	CreateMotionController(IsRightHand);
	if (CurrentMapName == FString("Start")) {//If it is in the main menu, automatically activate the VR pointer.
		EnableMenu();
	}
	else {
		//Hide the VR Pointer and show the hand
		HandMesh->SetVisibility(true, false);
		MotionControllerMesh->SetVisibility(false, true);
		if (WidgetInteractionComponent) {
			WidgetInteractionComponent->GetOwner()->SetActorHiddenInGame(true);
		}
		
	}
}

// Called every frame
void ATP_MotionController::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	UpdateHandAnimation();
}

AActor* ATP_MotionController::GetActorNearHand() const
{
	AActor* NearestOverlappingActor = nullptr;
	float NearestOverlap = 10000.0f;

	 // Filtered to StaticMeshes found in the level.
	TArray<AActor*> OverlapActors;
	GrabShpere->GetOverlappingActors(OverlapActors);
	FVector GrabSphereLocation = GrabShpere->GetComponentLocation();

	for (AActor* Actors : OverlapActors)
	{
		 // Filter to Actors that implement our interface for pickup/dropping
		bool bHasPickupInterface = Actors->GetClass()->ImplementsInterface(UTP_InteractionInterface::StaticClass());

		 // We want to only grab simulated meshes.
		if (bHasPickupInterface)
		{
			float MyLength = (Actors->GetActorLocation() - GrabSphereLocation).Size();
			// float MyLengthSquared = (Actors->GetActorLocation() - GrabSphereLocation).SizeSquared();

			if (MyLength < NearestOverlap)
			{
				NearestOverlappingActor = Actors;
				NearestOverlap = MyLength;
			}
		}
	}

	return NearestOverlappingActor;
}

void ATP_MotionController::GrabActor()
{
	bWantsToGrip = true;

	AActor* NearActor = GetActorNearHand();

	if (NearActor && NearActor->IsValidLowLevel() && !NearActor->IsPendingKill())
	{
		AttachedActor = NearActor;
		ITP_InteractionInterface::Execute_Pickup(NearActor, MotionController);
		RumbleController(0.2f);
	}
}

void ATP_MotionController::ReleaseActor()
{
	bWantsToGrip = false;

	if (AttachedActor && AttachedActor->IsValidLowLevel() && !AttachedActor->IsPendingKill())
	{
		 // Make sure this hand is still holding the Actor (May have been taken by another hand / event)
		if (AttachedActor->GetRootComponent()->GetAttachParent() == MotionController)
		{
			if (AttachedActor->GetClass()->ImplementsInterface(UTP_InteractionInterface::StaticClass()))
			{
				ITP_InteractionInterface::Execute_Drop(AttachedActor); // This is the Execute_* function. The asterisk means your function name. :)

				RumbleController(0.05f);

				 // Clear the reference
				AttachedActor = nullptr;
			}
		}
		else
		{
			 // Clear the reference
			AttachedActor = nullptr;
		}
	}
}


 // Rumble Controller
void ATP_MotionController::RumbleController(float Intensity)
{
	// GetPlayerController
	APlayerController* MyPlayerController = GetWorld()->GetFirstPlayerController(); // = UGameplayStatics::GetPlayerController(GetWorld(), 0);

	if (UHeadMountedDisplayFunctionLibrary::IsHeadMountedDisplayEnabled())
	{
		MyPlayerController->PlayHapticEffect(VRHapticEffect, Hand, Intensity);
	}
	else
	{
		// GamePad ForceFeedback, Deprecated in Epic Template
		MyPlayerController->PlayDynamicForceFeedback(Intensity, 0.1f, true, true, true, true, EDynamicForceFeedbackAction::Start);
	}
}

 // Rumble Controller when overlapping valid StaticMesh
void ATP_MotionController::OnComponentBeginOverlap(UPrimitiveComponent* OverlappedComp, AActor* OtherActor, UPrimitiveComponent* OtherComp, int32 OtherBodyIndex, bool bFromSweep, const FHitResult& SweepResult)
{
	if ((OtherComp != nullptr) && (OtherComp != GrabShpere))
	{
		// Cast the OverlapComponet to UStaticMeshComponent
		UStaticMeshComponent* const MyOverlapComponent = Cast<UStaticMeshComponent>(OtherComp);

		if (MyOverlapComponent && MyOverlapComponent->IsSimulatingPhysics())
		{
			RumbleController(0.05f);
		}
	}
}

void ATP_MotionController::UpdateHandAnimation()
{
	 // Update Animation of Hand
	if (AttachedActor || bWantsToGrip)
	{
		CurrentGripState = EGrip_Code::Grab;
	}
	else
	{
		AActor* NearActor = GetActorNearHand();

		if (NearActor)
		{
			CurrentGripState = EGrip_Code::CanGrab;
		}
		else
		{
			if (bWantsToGrip)
			{
				CurrentGripState = EGrip_Code::Grab;
			}
			else
			{
				CurrentGripState = EGrip_Code::Open;
			}
		}
	}

	 // Update the animation state of the hand mesh.
	UTP_HandAnimInstance* HandAnimation = Cast<UTP_HandAnimInstance>(HandMesh->GetAnimInstance());
	HandAnimation->SetGripState(CurrentGripState);

	 // Only let hand collide with environment while gripping
	// Not Use Switch
	if (CurrentGripState == EGrip_Code::Open || CurrentGripState == EGrip_Code::CanGrab)
	{
		HandMesh->SetCollisionEnabled(ECollisionEnabled::NoCollision);
	}
	else if (CurrentGripState == EGrip_Code::Grab)
	{
		HandMesh->SetCollisionEnabled(ECollisionEnabled::QueryAndPhysics);
	}
}

bool ATP_MotionController::GetIsRightHand()
{
	return Hand == EControllerHand::Right;
	
}

void ATP_MotionController::TriggerStart()
{
}

void ATP_MotionController::TriggerEnd()
{
}

void ATP_MotionController::GetMotionControllerMeshes()
{
	static ConstructorHelpers::FObjectFinder<UStaticMesh> SM_OculusLeft(TEXT("/Game/Meshes/OculusLeft"));
	if (SM_OculusLeft.Succeeded())
		LeftControllerMesh = SM_OculusLeft.Object;

	static ConstructorHelpers::FObjectFinder<UStaticMesh> SM_OculusRight(TEXT("/Game/Meshes/OculusRight"));
	if (SM_OculusRight.Succeeded())
		RightControllerMesh = SM_OculusRight.Object;
}

void ATP_MotionController::CreateMotionController(bool IsRightHand)
{
	
	if (IsRightHand) {
		if (RightControllerMesh) {
			MotionControllerMesh->SetStaticMesh(RightControllerMesh);
			MotionControllerMesh->SetRelativeRotation(FRotator(90, 315, 0));
		}
	}
	else {
		if (LeftControllerMesh) {
			MotionControllerMesh->SetStaticMesh(LeftControllerMesh);
			MotionControllerMesh->SetRelativeRotation(FRotator(90, 315, 0));
		}
	}
	
	
	if (MotionControllerMesh->GetStaticMesh()) {
		MotionController->SetWorldScale3D(FVector(1, 1, 1));
		
		MotionControllerMesh->SetRelativeLocation(FVector(16.1, -5.7, IsRightHand?.75:-.75));
		
		//
	}
	
}

void ATP_MotionController::SpawnWidget()
{
	bool IsRightHand = (Hand == EControllerHand::Right);

	
	
	UWorld* const World = GetWorld();
	if (World&&IsRightHand) {
		//Check for valid pointer
		if (!ensure(WidgetInteractionBlueprint))
		{
			UE_LOG(LogTemp, Warning, TEXT("Please set a widget interaction componenet (the pointer thing that allows you to click). This is why the ensure failed."));
			return;
		}

		//in world and right hand.
		AActor* WidgetActor = World->SpawnActor<AActor>(WidgetInteractionBlueprint);
		if (WidgetActor) {
			//made actor
			WidgetActor->SetActorTransform(HandMesh->GetComponentTransform());
			//WidgetActor->FinishSpawning(HandMesh->GetComponentTransform());
			WidgetInteractionComponent = Cast< UWidgetInteractionComponent>(WidgetActor->GetRootComponent());

			WidgetInteractionComponent->AttachToComponent(MotionControllerMesh, FAttachmentTransformRules::SnapToTargetNotIncludingScale);
			WidgetInteractionComponent->SetWorldLocation(MotionControllerMesh->GetComponentLocation());
			WidgetInteractionComponent->AddRelativeLocation(FVector(0, 1.6, 4.5));
			WidgetInteractionComponent->SetRelativeRotation(FRotator(-45, 90,0));
			WidgetInteractionComponent->PointerIndex = 1;
			
		}
	}
}



void ATP_MotionController::SetOtherController(UStaticMeshComponent * OtherControllerIn)
{
	OtherController = OtherControllerIn;
}

UStaticMeshComponent * ATP_MotionController::GetOtherController()
{
	return OtherController;
}

UStaticMeshComponent * ATP_MotionController::GetMotionControllerMesh()
{
	return MotionControllerMesh;
}

void ATP_MotionController::SetWidgetInteractionBlueprint(TSubclassOf<AActor> NewWidgetInteractionBlueprint)
{
	WidgetInteractionBlueprint = NewWidgetInteractionBlueprint;
}

void ATP_MotionController::EnableMenu()
{
	//Show the pointer and the controller model.
	HandMesh->SetVisibility(false, false);
	MotionControllerMesh->SetVisibility(true,true);
	HandMesh->SetWorldScale3D(FVector(1.0f, 1.0f, 1.0f));
	if (WidgetInteractionComponent) {
		WidgetInteractionComponent->GetOwner()->SetActorHiddenInGame(false);
	}
	else {
		SpawnWidget();
	}
}

void ATP_MotionController::DisableMenu()
{
	//Hide the pointer and the controller model.
	HandMesh->SetVisibility(true, false);
	MotionControllerMesh->SetVisibility(false, true);
	HandMesh->SetWorldScale3D(FVector(1.0f, 1.0f, 1.0f));
	bool IsRightHand = (Hand == EControllerHand::Right);
	if (!IsRightHand) {
		HandMesh->SetWorldScale3D(FVector(1.0f, 1.0f, -1.0f));
	}
	if (WidgetInteractionComponent) {
		WidgetInteractionComponent->GetOwner()->SetActorHiddenInGame(true);
	}
}
